package mezz.jei.common.input;

import mezz.jei.common.input.handlers.CombinedInputHandler;
import mezz.jei.core.util.ReflectionUtil;
import net.minecraft.class_310;
import net.minecraft.class_342;
import net.minecraft.class_437;
import java.util.List;

public class ClientInputHandler {
	private final List<ICharTypedHandler> charTypedHandlers;
	private final CombinedInputHandler inputHandler;
	private final IInternalKeyMappings keybindings;
	private final ReflectionUtil reflectionUtil = new ReflectionUtil();

	public ClientInputHandler(List<ICharTypedHandler> charTypedHandlers, CombinedInputHandler inputHandler, IInternalKeyMappings keybindings) {
		this.charTypedHandlers = charTypedHandlers;
		this.inputHandler = inputHandler;
		this.keybindings = keybindings;
	}

	public void onInitGui() {
		this.inputHandler.handleGuiChange();
	}

	/**
	 * When we have keyboard focus, use Pre
	 */
	public boolean onKeyboardKeyPressedPre(class_437 screen, UserInput input) {
		if (!isContainerTextFieldFocused(screen)) {
			return this.inputHandler.handleUserInput(screen, input, keybindings)
				.isPresent();
		}
		return false;
	}

	/**
	 * Without keyboard focus, use Post
	 */
	public boolean onKeyboardKeyPressedPost(class_437 screen, UserInput input) {
		if (isContainerTextFieldFocused(screen)) {
			return this.inputHandler.handleUserInput(screen, input, keybindings)
				.isPresent();
		}
		return false;
	}

	/**
	 * When we have keyboard focus, use Pre
	 */
	public boolean onKeyboardCharTypedPre(class_437 screen, char codePoint, int modifiers) {
		if (!isContainerTextFieldFocused(screen)) {
			return handleCharTyped(codePoint, modifiers);
		}
		return false;
	}

	/**
	 * Without keyboard focus, use Post
	 */
	public boolean onKeyboardCharTypedPost(class_437 screen, char codePoint, int modifiers) {
		if (isContainerTextFieldFocused(screen)) {
			return handleCharTyped(codePoint, modifiers);
		}
		return false;
	}

	public boolean onGuiMouseClicked(class_437 screen, UserInput input) {
		boolean handled = this.inputHandler.handleUserInput(screen, input, keybindings)
			.isPresent();

		if (class_310.method_1551().field_1755 == screen && input.is(keybindings.getLeftClick())) {
			handled |= this.inputHandler.handleDragStart(screen, input)
				.isPresent();
		}
		return handled;
	}

	public boolean onGuiMouseReleased(class_437 screen, UserInput input) {
		boolean handled = this.inputHandler.handleUserInput(screen, input, keybindings)
			.isPresent();

		if (input.is(keybindings.getLeftClick())) {
			handled |= this.inputHandler.handleDragComplete(screen, input)
				.isPresent();
		}
		return handled;
	}

	public boolean onGuiMouseScroll(double mouseX, double mouseY, double scrollDelta) {
		return this.inputHandler.handleMouseScrolled(mouseX, mouseY, scrollDelta);
	}

	private boolean handleCharTyped(char codePoint, int modifiers) {
		return this.charTypedHandlers.stream()
			.filter(ICharTypedHandler::hasKeyboardFocus)
			.anyMatch(handler -> handler.onCharTyped(codePoint, modifiers));
	}

	private boolean isContainerTextFieldFocused(class_437 screen) {
		return reflectionUtil.getFieldWithClass(screen, class_342.class)
			.anyMatch(textField -> textField.method_37303() && textField.method_25370());
	}
}
